@@ -26,24 +26,31 @@ import subprocess
 import shutil
 from select import select
 from contextlib import contextmanager
-from copy import copy
 from functools import partial
+if PY2:
+    import __builtin__ as builtins
+else:
+    import builtins
 
 from coconut.terminal import (
     logger,
     complain,
+    internal_assert,
 )
 from coconut.exceptions import (
     CoconutException,
     get_encoding,
-    internal_assert,
 )
 from coconut.constants import (
+    WINDOWS,
+    PY34,
+    PY32,
     fixpath,
+    base_dir,
     main_prompt,
     more_prompt,
     default_style,
-    default_histfile,
+    prompt_histfile,
     prompt_multiline,
     prompt_vi_mode,
     prompt_wrap_lines,
@@ -53,19 +60,17 @@ from coconut.constants import (
     tutorial_url,
     documentation_url,
     reserved_vars,
-    num_added_tb_layers,
     minimum_recursion_limit,
     oserror_retcode,
     base_stub_dir,
     installed_stub_dir,
-    WINDOWS,
-    PY34,
-    PY32,
+    interpreter_uses_auto_compilation,
+    interpreter_uses_coconut_breakpoint,
 )
 
 if PY26:
     import imp
-if not PY26:
+else:
     import runpy
 try:
     # just importing readline improves built-in input()
@@ -97,13 +102,13 @@ except ImportError:
 except KeyError:
     complain(
         ImportError(
-            "detected outdated pygments version (run 'pip install --upgrade pygments' to fix)",
+            "detected outdated pygments version (run '{python} -m pip install --upgrade pygments' to fix)".format(python=sys.executable),
         ),
     )
     prompt_toolkit = None
 
 # -----------------------------------------------------------------------------------------------------------------------
-# FUNCTIONS:
+# UTILITIES:
 # -----------------------------------------------------------------------------------------------------------------------
 
 
@@ -163,9 +168,8 @@ def rem_encoding(code):
 def exec_func(code, glob_vars, loc_vars=None):
     """Wrapper around exec."""
     if loc_vars is None:
-        exec(code, glob_vars)
-    else:
-        exec(code, glob_vars, loc_vars)
+        loc_vars = glob_vars
+    exec(code, glob_vars, loc_vars)
 
 
 def interpret(code, in_vars):
@@ -191,7 +195,8 @@ def handling_broken_process_pool():
         try:
             yield
         except BrokenProcessPool:
-            raise KeyboardInterrupt()
+            logger.log_exc()
+            raise KeyboardInterrupt("broken process pool")
 
 
 def kill_children():
@@ -201,7 +206,7 @@ def kill_children():
     except ImportError:
         logger.warn(
             "missing psutil; --jobs may not properly terminate",
-            extra="run 'pip install coconut[jobs]' to fix",
+            extra="run '{python} -m pip install coconut[jobs]' to fix".format(python=sys.executable),
         )
     else:
         parent = psutil.Process()
@@ -292,8 +297,17 @@ def run_cmd(cmd, show_output=True, raise_errs=True, **kwargs):
 
 def symlink(link_to, link_from):
     """Link link_from to the directory link_to universally."""
-    if os.path.exists(link_from) and not os.path.islink(link_from):
-        shutil.rmtree(link_from)
+    if os.path.exists(link_from):
+        if os.path.islink(link_from):
+            os.unlink(link_from)
+        elif WINDOWS:
+            try:
+                os.rmdir(link_from)
+            except OSError:
+                logger.log_exc()
+                shutil.rmtree(link_from)
+        else:
+            shutil.rmtree(link_from)
     try:
         if PY32:
             os.symlink(link_to, link_from, target_is_directory=True)
@@ -301,25 +315,35 @@ def symlink(link_to, link_from):
             os.symlink(link_to, link_from)
     except OSError:
         logger.log_exc()
-    else:
-        return
     if not os.path.islink(link_from):
         shutil.copytree(link_to, link_from)
 
 
+def install_mypy_stubs():
+    """Properly symlink mypy stub files."""
+    symlink(base_stub_dir, installed_stub_dir)
+    return installed_stub_dir
+
+
+def set_env_var(name, value):
+    """Universally set an environment variable."""
+    os.environ[py_str(name)] = py_str(value)
+
+
 def set_mypy_path():
     """Put Coconut stubs in MYPYPATH."""
-    symlink(base_stub_dir, installed_stub_dir)
+    install_dir = install_mypy_stubs()
     original = os.environ.get(mypy_path_env_var)
     if original is None:
-        new_mypy_path = installed_stub_dir
-    elif not original.startswith(installed_stub_dir):
-        new_mypy_path = installed_stub_dir + os.pathsep + original
+        new_mypy_path = install_dir
+    elif not original.startswith(install_dir):
+        new_mypy_path = install_dir + os.pathsep + original
     else:
         new_mypy_path = None
     if new_mypy_path is not None:
-        os.environ[mypy_path_env_var] = new_mypy_path
-    logger.log_func(lambda: (mypy_path_env_var, "=", os.environ[mypy_path_env_var]))
+        set_env_var(mypy_path_env_var, new_mypy_path)
+    logger.log_func(lambda: (mypy_path_env_var, "=", os.environ.get(mypy_path_env_var)))
+    return install_dir
 
 
 def stdin_readable():
@@ -344,10 +368,11 @@ def set_recursion_limit(limit):
 
 
 def _raise_ValueError(msg):
+    """Raise ValueError(msg)."""
     raise ValueError(msg)
 
 
-def canparse(argparser, args):
+def can_parse(argparser, args):
     """Determines if argparser can parse args."""
     old_error_method = argparser.error
     argparser.error = _raise_ValueError
@@ -361,6 +386,27 @@ def canparse(argparser, args):
         argparser.error = old_error_method
 
 
+def subpath(path, base_path):
+    """Check if path is a subpath of base_path."""
+    path = fixpath(path)
+    base_path = fixpath(base_path)
+    return path == base_path or path.startswith(base_path + os.sep)
+
+
+def invert_mypy_arg(arg):
+    """Convert --arg into --no-arg or equivalent."""
+    if arg.startswith("--no-"):
+        return "--" + arg[len("--no-"):]
+    elif arg.startswith("--allow-"):
+        return "--disallow-" + arg[len("--allow-"):]
+    elif arg.startswith("--disallow-"):
+        return "--allow-" + arg[len("--disallow-"):]
+    elif arg.startswith("--"):
+        return "--no-" + arg[len("--"):]
+    else:
+        return None
+
+
 # -----------------------------------------------------------------------------------------------------------------------
 # CLASSES:
 # -----------------------------------------------------------------------------------------------------------------------
@@ -377,7 +423,8 @@ class Prompt(object):
         """Set up the prompt."""
         if prompt_toolkit is not None:
             self.set_style(os.environ.get(style_env_var, default_style))
-            self.set_history_file(default_histfile)
+            self.set_history_file(prompt_histfile)
+            self.lexer = PygmentsLexer(CoconutLexer)
 
     def set_style(self, style):
         """Set pygments syntax highlighting style."""
@@ -414,7 +461,7 @@ class Prompt(object):
             except EOFError:
                 raise  # issubclass(EOFError, Exception), so we have to do this
             except (Exception, AssertionError):
-                logger.display_exc()
+                logger.print_exc()
                 logger.show_sig("Syntax highlighting failed; switching to --style none.")
                 self.style = None
         return input(msg)
@@ -433,7 +480,7 @@ class Prompt(object):
             vi_mode=self.vi_mode,
             wrap_lines=self.wrap_lines,
             enable_history_search=self.history_search,
-            lexer=PygmentsLexer(CoconutLexer),
+            lexer=self.lexer,
             style=style_from_pygments_cls(
                 pygments.styles.get_style_by_name(self.style),
             ),
@@ -446,10 +493,10 @@ class Runner(object):
     def __init__(self, comp=None, exit=sys.exit, store=False, path=None):
         """Create the executor."""
         from coconut.convenience import auto_compilation, use_coconut_breakpoint
-        auto_compilation(on=True)
-        use_coconut_breakpoint(on=False)
+        auto_compilation(on=interpreter_uses_auto_compilation)
+        use_coconut_breakpoint(on=interpreter_uses_coconut_breakpoint)
         self.exit = exit
-        self.vars = self.build_vars(path)
+        self.vars = self.build_vars(path, init=True)
         self.stored = [] if store else None
         if comp is not None:
             self.store(comp.getheader("package:0"))
@@ -457,7 +504,7 @@ class Runner(object):
             self.fix_pickle()
 
     @staticmethod
-    def build_vars(path=None):
+    def build_vars(path=None, init=False):
         """Build initial vars."""
         init_vars = {
             "__name__": "__main__",
@@ -466,9 +513,12 @@ class Runner(object):
         }
         if path is not None:
             init_vars["__file__"] = fixpath(path)
-        # put reserved_vars in for auto-completion purposes
-        for var in reserved_vars:
-            init_vars[var] = None
+        if init:
+            # put reserved_vars in for auto-completion purposes only at the very beginning
+            for var in reserved_vars:
+                # but don't override any default Python built-ins
+                if var not in dir(builtins):
+                    init_vars[var] = None
         return init_vars
 
     def store(self, line):
@@ -481,7 +531,10 @@ class Runner(object):
         from coconut import __coconut__  # this is expensive, so only do it here
         for var in self.vars:
             if not var.startswith("__") and var in dir(__coconut__):
-                self.vars[var] = getattr(__coconut__, var)
+                cur_val = self.vars[var]
+                static_val = getattr(__coconut__, var)
+                if getattr(cur_val, "__doc__", None) == getattr(static_val, "__doc__", None):
+                    self.vars[var] = static_val
 
     @contextmanager
     def handling_errors(self, all_errors_exit=False):
@@ -492,8 +545,8 @@ class Runner(object):
             self.exit(err.code)
         except BaseException:
             etype, value, tb = sys.exc_info()
-            for _ in range(num_added_tb_layers):
-                if tb is None:
+            while True:
+                if tb is None or not subpath(tb.tb_frame.f_code.co_filename, base_dir):
                     break
                 tb = tb.tb_next
             traceback.print_exception(etype, value, tb)
@@ -518,6 +571,7 @@ class Runner(object):
             run_func = eval
         else:
             run_func = exec_func
+        result = None
         with self.handling_errors(all_errors_exit):
             if path is None:
                 result = run_func(code, self.vars)
@@ -529,7 +583,7 @@ class Runner(object):
                     self.vars.update(use_vars)
             if store:
                 self.store(code)
-            return result
+        return result
 
     def run_file(self, path, all_errors_exit=True):
         """Execute a Python file."""
@@ -551,15 +605,18 @@ class Runner(object):
 
 class multiprocess_wrapper(object):
     """Wrapper for a method that needs to be multiprocessed."""
+    __slots__ = ("rec_limit", "logger", "argv", "base", "method")
 
     def __init__(self, base, method):
         """Create new multiprocessable method."""
-        self.recursion = sys.getrecursionlimit()
-        self.logger = copy(logger)
+        self.rec_limit = sys.getrecursionlimit()
+        self.logger = logger.copy()
+        self.argv = sys.argv
         self.base, self.method = base, method
 
     def __call__(self, *args, **kwargs):
         """Call the method."""
-        sys.setrecursionlimit(self.recursion)
+        sys.setrecursionlimit(self.rec_limit)
         logger.copy_from(self.logger)
+        sys.argv = self.argv
         return getattr(self.base, self.method)(*args, **kwargs)
